diff --git a/src/data.c b/src/data.c
index ea611ad1abf..a3fba48c7f2 100644
--- a/src/data.c
+++ b/src/data.c
@@ -2570,6 +2570,33 @@ DEFUN ("aref", Faref, Saref, 2, 2, 0,
     }
 }
 
+static void
+warn_aset_string_resize (Lisp_Object string, const char *msg, unsigned ch)
+{
+  /* Only warn the first few times to avoid spamming the user.  */
+  static int remaining = 5;
+  if (remaining <= 0)
+    return;
+  --remaining;
+
+  Lisp_Object bt[16];
+  get_backtrace (bt, ARRAYELTS (bt));
+  Lisp_Object warning
+    = CALLN (Fformat,
+	     build_string ("`aset' on %s string: %s;"
+			   " this will become an error in Emacs 31."
+			   "  Call chain (most recent first):"),
+	     build_string (STRING_MULTIBYTE (string) ? "multibyte" : "unibyte"),
+	     CALLN (Fformat, build_string (msg), make_int (ch)));
+
+  for (ptrdiff_t i = EQ (bt[0], Qaset) ? 1 : 0;
+       i < ARRAYELTS (bt) && !NILP (bt[i]); i++)
+    if (SYMBOLP (bt[i]))
+      warning = CALLN (Fformat, build_string ("%s %s;"), warning, bt[i]);
+  Vdelayed_warnings_list = Fcons (list2 (Qaset, warning),
+				  Vdelayed_warnings_list);
+}
+
 DEFUN ("aset", Faset, Saset, 3, 3, 0,
        doc: /* Store into the element of ARRAY at index IDX the value NEWELT.
 Return NEWELT.  ARRAY may be a vector, a string, a char-table or a
@@ -2623,6 +2650,12 @@ DEFUN ("aset", Faset, Saset, 3, 3, 0,
 	{
 	  idxval_byte = string_char_to_byte (array, idxval);
 	  p1 = SDATA (array) + idxval_byte;
+	  if (c > 127)
+	    warn_aset_string_resize (array,
+				     "replacement char #x%x is non-ASCII", c);
+	  else if (*p1 > 127)
+	    warn_aset_string_resize (array,
+				     "replaced char #x%x is non-ASCII", *p1);
 	  prev_bytes = BYTES_BY_CHAR_HEAD (*p1);
 	}
       else if (SINGLE_BYTE_CHAR_P (c))
@@ -2632,6 +2665,9 @@ DEFUN ("aset", Faset, Saset, 3, 3, 0,
 	}
       else
 	{
+	  warn_aset_string_resize (array,
+				   "replacement char #x%x is not single-byte",
+				   c);
 	  for (ptrdiff_t i = SBYTES (array) - 1; i >= 0; i--)
 	    if (!ASCII_CHAR_P (SREF (array, i)))
 	      args_out_of_range (array, newelt);
